/******************************************************************************
**                                                                           **
**    ~elderK's portable AVL implementation                                  **
**  ---------------------------------------------                            **
**                                                                           **
**    Source: avl.h                                                          **
**                                                                           **
**    Purpose:                                                               **
**      This file defines ~kAVL's public interface.                          **
**                                                                           **
**      This code is self contained. External software is expected to handle **
**      the allocation and deallocation of nodes, as well as the generation  **
**      keys.                                                                **
**                                                                           **
**    Author: DG.Ward (aka: ~elderK)                                         **
**                                                                           **
**    Contact:                                                               **
**      warddg1@tekotago.ac.nz                                               **
**      kyretzn@gmail.com                                                    **
**      comeherenightmare@hotmail.com                                        **
**                                                                           **
**    Date: December 10, 2007.                                               **
**                                                                           **
******************************************************************************/
#ifndef _h_K_avl
#define _h_K_avl /* _h_K_avl */

/* -- avl_hdr_t
 *
 * avl_hdr_t is the base structure for any AVL node.
 * It contains three pointers, obviously one to the
 * node's two children and to the node's direct
 * parent.
 *
 * Anyone familiar with trees will understand what
 * the key is for. It is responsible for:
 *  - maintaining the BST property.
 *  - providing a unique ID for lookup.
 *
 * nBal stores the current balance of the node,
 * it is used primarily for recognizing when a
 * tree is in need of rebalancing.
 *
 * nRel states the relationship between the node
 * and it's parent (left or right child). This
 * is included to reduce comparisons during
 * insertion and removal.
 *
 * If you are pedantic and wish to squeeze ONE
 * byte out of the structural size, it is
 * acceptible to place nBal and nRel in a bitfield
 * as follows:
 *
 *   unsigned char nBal  : 3,
 *                 nRel  : 2,
 *                 nResv : 3;
 *
 * nResv can store whatever you please...
 * If you *do* optimize for size, ensure you add
 * __attribute__((packed)). Also, note that
 * unsigned char bitfields are not ANSI C89
 * compliant.
 *
 * -- avl_hdr_t
*/
typedef
struct _avl_hdr_t
{
    struct _avl_hdr_t *nLinks[3];
    unsigned long int nKey;

    unsigned char nBal,
                  nRel;
}
avl_hdr_t;

/* -- init_node
 *
 * This call exists to initialize nodes, to make them
 * appropriate for tree insertion.
 *
 * Parameters:
 *   #1 - node, avl_hdr_t *
 *        Pointer to a valid avl_hdr_t structure.
 *        If this is null, the function will
 *        simply report failure.
 *
 *   #2 - key, unsigned long int.
 *        The identifier for the new node.
 *
 * Returns:
 *   On successful creation, this function will
 *   return a pointer to a new node.
 *
 *   On failure, it will return (avl_hdr_t *)0.
 *
 * --
*/
extern avl_hdr_t *init_node (avl_hdr_t *, unsigned long int);

/* -- insert_node
 *
 * This call will insert the node passed by the second
 * parameter, into the tree pointed to by the first.
 *
 * Please note, that the first paremeter is a POINTER
 * to a POINTER. This is so, if needed, the function can
 * update the 'tree' pointers target.
 *
 * Parameters:
 *  #1 - tree, avl_hdr_t **.
 *       A pointer to the root node of the tree.
 *       The target of the pointer may be NULL.
 *       (No existing root node in tree).
 *
 *  #2 - node, avl_hdr_t *
 *       Pointer to the node destined to inhabit the
 *       the tree. Nodes scheduled for insertion MUST
 *       be initialized. Inserting an uninitialized node
 *       will result in undefined behaviour.
 *       (Read: Initialize the nodes, or your program will
 *       crash faster than Windows 95...)
 *
 * Returns:
 *  On successful insertion, a pointer to the node inserted
 *  will be returned;
 *
 *  (avl_hdr_t *)0 will be returned in the case of
 *  key collision.
 *
 * Example of use:
 *
 *  avl_hdr_t *my_new_tree = (avl_hdr_t *)0;
 *  avl_hdr_t *a_node = create_node (0xCAFEBABE);
 *
 *  if (insert_node (&my_new_tree, a_node) == (avl_hdr_t *)0)
 *  {
 *    free (a_node);
 *    ... do error stuff ...
 *  }
 *
 * --
*/
extern avl_hdr_t *insert_node (avl_hdr_t **, avl_hdr_t *);

/* -- locate_node
 *
 * This call will attempt to locate the node identified
 * by the key provided.
 *
 * Parameters:
 *  #1 - tree, avl_hdr_t **
 *       A pointer to the root node of the tree.
 *       The target of the pointer may be NULL.
 *
 *  #2 - key, unsigned long int
 *       Idenfitier of that evasive, evil node...
 *       (Where in the world is Carmen Sandiego?)
 *
 * Returns:
 *  On success, a pointer to the now-discovered node is
 *  returned.
 *
 *  In all other cases, (avl_hdr_t *)0 is returned.
 *
 * --
*/
extern avl_hdr_t *locate_node (avl_hdr_t **, unsigned long int);

/* -- delete_node
 *
 * This call will remove the node identified from the
 * specified tree. Note, that this function may NOT release
 * the node you pass, due to the details of how this function
 * works.
 *
 * This call does NOT take responsibility for the deallocation
 * of the desired node. This makes it possible to use this call
 * to transfer nodes between trees.
 *
 * Parameters:
 *  #1 - tree, avl_hdr_t **
 *       A pointer to the root node of the tree.
 *       The target of the pointer may be NULL.
 *
 *  #2 - node, avl_hdr_t *
 *       Pointer to the node scheduled for deletion.
 *       This can only point to null in the case that the tree
 *       points to NULL. In any other case, this must NOT
 *       be zeroed.
 *
 * Returns:
 *  pointer to a node containing the data of the node removed.
 *
 *  For those anxious, no, this does not return a copy or have
 *  anything to do with wasting memory :). It is simply a side
 *  effect of implementation detail.
 *
 * --
*/
extern avl_hdr_t *delete_node (avl_hdr_t **, avl_hdr_t *);

#endif /* h_K_avl */

