/******************************************************************************
**                                                                           **
**    ~elderK's portable AVL implementation                                  **
**  ---------------------------------------------                            **
**                                                                           **
**    Source: avl.c                                                          **
**                                                                           **
**    Purpose:                                                               **
**      This file contains all of ~kAVL's private implementation details.    **
**                                                                           **
**      This code is self contained. External software is expected to handle **
**      the allocation and deallocation of nodes, as well as the generation  **
**      keys.                                                                **
**                                                                           **
**    Author: DG.Ward (aka: ~elderK)                                         **
**                                                                           **
**    Contact:                                                               **
**      warddg1@tekotago.ac.nz                                               **
**      kyretzn@gmail.com                                                    **
**      comeherenightmare@hotmail.com                                        **
**                                                                           **
**    Date: December 10, 2007.                                               **
**                                                                           **
******************************************************************************/
#include <avl.h>

/* -- avl_trace_en
 *
 * Mode select for trace (see below).
 *
 * --
*/
enum
{
    INSERTION,
    DELETION
}
avl_trace_en;

/* -- avl_balance_en
 *
 * Balance definitions, suited for
 * 3bit representation.
 *
 * LC = 0
 * LH = 1
 * B  = 2
 * RH = 3
 * RC = 4
 *
 * --
*/
enum
{
    LEFT_CRITICAL,
    LEFT_HEAVY,
    BALANCED,
    RIGHT_HEAVY,
    RIGHT_CRITICAL
}
avl_balance_en;

/* -- avl_relationship_en
 *
 * These define a node's relationship to it's
 * parent. INIT_CHILD is ONLY present in a
 * newly initialized node. In any other case,
 * INIT_CHILD signals some kind of catastrophic
 * failure.
 *
 * --
*/
enum
{
    LEFT_CHILD,
    RIGHT_CHILD,
    INIT_CHILD
}
avl_relationship_en;

/* -- avl_links_en
 *
 * Friendly definitions for nLink subscripts :)
 *
 * --
*/
enum
{
    LEFT,
    RIGHT,
    PARENT
}
avl_links_en;

/* -- avl_classification_en
 *
 * Friendly definitions for the various node
 * types.
 *
 * --
*/
enum
{
    LEAF,
    BRANCH,
    SUBTREE
}
avl_classification_en;

/* -- avl constants
 *
 * The following are a set of constants, used simply to reduce the
 * amount of comparisons at runtime.
 *
 * To understand why these constants are what they are, refer to a
 * reference describing the AVL data structure and, follow the
 * rebalance_subtree routine with vigilance! (and coffee, lots of coffee).
 *
 * --
*/
unsigned char direct_constants[2] =
{
    RIGHT_HEAVY,
    LEFT_HEAVY
};

signed char trace_constants[2] =
{
    -1,
    +1
};

unsigned char indirect_constants[2][2] =
{
    {LEFT_HEAVY, RIGHT_HEAVY},
    {RIGHT_HEAVY, LEFT_HEAVY}
};

unsigned char indirect_balances[2][4] =
{
    {BALANCED, RIGHT_HEAVY, LEFT_HEAVY, BALANCED},
    {BALANCED, LEFT_HEAVY, RIGHT_HEAVY, BALANCED}
};
/*     /|__________________________________|\
                         |
                         V
        Subscripts 0 and 3 are not neccessary.
       Mental note: Prune them at a later date.
*/

/*
 * -- rotate_node
 *
 * This call rotates the subtree specified by target_node,
 * DIRECTION around target_node. This call also takes care
 * of updating the TREE's target, as well as ensuring that
 * children are not lost. :) No one likes orphans!!!
 *
 * Parameters:
 *  #1 - tree, avl_hdr_t **
 *       Pointer to a pointer targetting the tree's root node.
 *  #2 - target_node, avl_hdr_t *
 #       Pointer to the node/subtree to be rotated.
 #  #3 - direction, unsigned char
 #       Direction to rotate the node/subtree.
 #
 # Returns:
 #  A pointer to the INVERSE-DIRECTION node of target_node.
 #  (ie: The subtree's new root...)
 #
 * --
*/
static
avl_hdr_t
*rotate_node (avl_hdr_t **tree,
              avl_hdr_t *target_node,
              unsigned char direction)
{
    avl_hdr_t *successor = target_node->nLinks[!direction];

    target_node->nLinks[!direction] = successor->nLinks[direction];
    successor->nLinks[direction] = target_node;

    if (target_node->nLinks[PARENT])
        successor->nLinks[PARENT] = target_node->nLinks[PARENT];
    else
    {
        successor->nLinks[PARENT] = (avl_hdr_t *)0;
        *tree = successor;
    }

    successor->nRel = target_node->nRel;
    target_node->nLinks[PARENT] = successor;
    target_node->nRel = direction;

    if (target_node->nLinks[!direction])
    {
        target_node->nLinks[!direction]->nRel = !direction;
        target_node->nLinks[!direction]->nLinks[PARENT] = target_node;
    }

    return successor;
}

/* -- rebalance_subtree
 *
 * This call carries the heavy burden of resituating offending
 * nodes (oh, those bad, evil, nefarious critically imbalanced
 * nodes! Someone give them soem antidepressants, jeez!).
 *
 * The operation of this function depends entirely on the node
 * it is working on, as all balance-updates and rotations are
 * are based on the offending node's imbalance.
 *
 * To get any serious understanding here, it really pays to
 * be familiar with the AVL data structure.
 *
 * The function rotates the tree as necessary, ensuring
 * all balances fall within the AVL property.
 *
 * Parameters:
 *  #1 - tree, avl_hdr_t **
 *       Pointer to a pointer targetting the root node.
 *  #2 - local_root, avl_hdr_t *
 *       Pointer to the critically imbalanced node.
 *
 * Returns:
 *  A pointer to a now balanced subtree.
 *
 * --
*/
static
avl_hdr_t
*rebalance_subtree (avl_hdr_t **tree,
                    avl_hdr_t *local_root)
{
    /* Is it a Left or Right imbalance?
       0 = Left imbalance,
       1 = Right imbalance. */
    unsigned char root_imbalance = local_root->nBal > BALANCED;

    /* Connect us to local_root's direct child (ie: what will
       become the new root of this subtree, taking local_root's
       place).

       The direct child is always the IMBALANCE child of local_root.
       Ie: Left imbalance, Direct child is local_root's left child. */
    avl_hdr_t *direct_child = local_root->nLinks[root_imbalance];

    /* Left Imbalance:
         Is direct_child's balance right heavy? (Left-right case).

       Right Imbalance:
         Is direct-child's balance left heavy? (Right-left case). */
    if (direct_child->nBal == direct_constants[root_imbalance])
    {
        /* Connect us to the indirect child. The indirect child is always
           direct_child's INVERSE-IMBALANCE child.

           For example:
             Left Imblance:
               direct_child = local_root->nLinks[LEFT]
               indirect_child = direct_child->nLinks[RIGHT]
        */
        avl_hdr_t *indirect_child =
            direct_child->nLinks[!root_imbalance];

        /* If LEFT_RIGHT case and, Indirect child has a left child.
           If RIGHT_LEFT case and, Indirect child has a right child. */
        if (indirect_child->nBal == BALANCED)
        {
            /* Left_Right_Left case:
               direct_child->nBal = Balanced
               local_root->nBal   = Right heavy

               Right_Left_Right case:
               direct_child->nBal = Balanced
               local_root->nBal   = Left heavy
            */
            direct_child->nBal = BALANCED;
            local_root->nBal = indirect_balances[root_imbalance][1];
        }
        else
        /* If LEFT_RIGHT case and, Indirect child has a right child.
           If RIGHT_LEFT case and, Indirect child has a left child. */
        if (indirect_child->nBal ==
            indirect_constants[root_imbalance][1])
        {
            /* Left_Right_Left case:
               direct_child->nBal = Left heavy
               local_root->nBal   = Balanced

               Right_Left_Right case:
               direct_child->nBal = Right heavy
               local_root->nBal   = Balanced
            */
            direct_child->nBal = indirect_balances[root_imbalance][2];
            local_root->nBal = BALANCED;
        }
        else
            /* LEFT_RIGHT_BALANCED, RIGHT_LEFT_BALANCED. */
            local_root->nBal = direct_child->nBal = BALANCED;

        /* The indirect child alway's ends up balanced. */
        indirect_child->nBal = BALANCED;

        /* If LEFT_IMBALANCE, Rotate direct_child left.
           If RIGHT_IMBALANCE, Rotate direct_child right. */
        local_root->nLinks[root_imbalance] =
            rotate_node (tree, direct_child, root_imbalance);
    }
    else
    if (direct_child->nBal == direct_constants[!root_imbalance])
    {
        /* Left-Left with a right child.
           Right-Right with a left child. */
        direct_child->nBal = BALANCED;
        local_root->nBal = BALANCED;
    }
    else
    if (direct_child->nBal == BALANCED)
    {
        /* Left-Left, no child.
             direct_child->nBal = Left heavy.
             local_child->nBal  = Right heavy.

           Right-Right, no child.
             direct_child->nBal = Right heavy.
             local_child->nBal  = Left heavy.
        */
        direct_child->nBal = direct_constants[root_imbalance];
        local_root->nBal = direct_constants[!root_imbalance];
    }

    /* If LEFT_IMBALANCE, Rotate local_root right.
       If RIGHT_IMBALANCE, Rotate local_root left. */
    return rotate_node (tree, local_root, !root_imbalance);
}

/* -- trace
 *
 * Trace exists only to maintain tree balance. It is called after
 * an Insertion or Deletion. It traces up towards the root, only
 * as necessary. There is no need to return to the root on all
 * occasions and the cases to which that applies will be explained
 * below.
 *
 * Parameters:
 *  #1 - tree, avl_hdr_t **
 *       A pointer to the tree's root. The target of the pointer to
 *       which tree points, must NOT be null if origin (see below)
 *       is live.
 *
 *  #2 - origin, avl_hdr_t *
 *       Points to the node from which to start tracing.
 *
 *  #3 - trace_mode, unsigned char
 *       Specifies whether the trace is following after an
 *       Insertion or Deletion.
 *
 * Returns:
 *  There is no return value.
 *
 * --
*/
static
void
trace (avl_hdr_t **tree,
       avl_hdr_t *origin,
       unsigned char trace_mode)
{
    /* Is the origin actually valid? It may be null in the case
       of a root-node deletion. */
    if (origin)
    {
        /* The origin is active, let's grab a link to it's parent! */
        avl_hdr_t *iterator = origin->nLinks[PARENT];

        /* We will use rel_cache to cache the origin's relationship,
           over trace iterations. */
        unsigned char rel_cache;

        /* The multiplier ensures that the balancing constants are
           appropriate for the mode of operation. */
        signed char multiplier = trace_mode ? -1 : 1;

        /* Trace until we reach a point where it is no longer
           necessary ... */
        while (1)
        {
            rel_cache = origin->nRel;

            /* Is the current origin node unbalanced? */
            switch (origin->nBal)
            {
                case LEFT_CRITICAL:
                case RIGHT_CRITICAL:
                {
                    /* Yes, it is! Does it have a valid parent? */
                    if (!iterator)
                        /* No, We are rebalancing the root node.
                           Ensure that the root pointer now targets
                           origin's successor. */
                        *tree = rebalance_subtree (tree, origin);
                    else
                    {
                        /* Yes, ensure origin's parent will point to
                           origin's successor. */
                        iterator->nLinks[rel_cache] =
                            rebalance_subtree (tree, origin);

                        /* This is one of the cases that a full retrace to the
                           root node is not neccesary. This is the case:
                           (Balances are in brackets, below the keys).

                                            32
                                          ( +1 )         .----- Critical Right.
                                16                      128
                              ( -1 )                  ( +2 )
                        8                       64              256
                      ( -1 )                   ( 0 )           ( +1 )
                      4                                     192     512
                    ( 0 )                                  ( 0 )   ( +1 )
                                                                       1024
                                                                       ( 0 )

                           In this case, We have removed 64's left child, a
                           node with the key 48. Before hand, node 128 was only
                           slightly right heavy. However, with the removal of
                           48, it becomes right critical. As is the case with a
                           right imbalance, we rotate the offending node left.
                           The result is as follows:

                                            32
                                          ( +1 )
                                16                      256
                              ( -1 )                  ( +1 )
                        8                       64              512
                      ( -1 )                  ( +1 )          ( +1 )
                      4                            192             1024
                    ( 0 )                         ( 0 )            ( 0 )

                           As you can see, the successor of 128, 256 is not
                           completely balanced. However, the tree is both
                           correct (BST property) and unchanged in height.
                           As such, there is no need to directly trace to the
                           root node.

                           This shortcut is valid at any position within a tree
                           exhibiting this case, regardless of tree size. As
                           such, it can save the program considerable amounts of
                           trace time :). */

                        if (trace_mode &&
                            iterator->nLinks[rel_cache]->nBal !=
                            BALANCED)
                            return;
                    }
                }

                case BALANCED:
                {
                    /* This is another case where we can save time, depending
                       on trace mode. If we are performing a trace after
                       an insertion, we know that the inserted node's parent
                       signals a height change when it is no longer completely
                       balanced.

                       Since tree height has not changed, there is no purpose
                       to perform a full retrace. */

                    if (trace_mode)
                        break;

                    return;
                }

                default:
                {
                    /* Once again in deletion, there is a shortcut! This
                       shortcut is useful, as it decides whether or not we have
                       to continue tracing at all.

                       The world of deletion is black and white. And so,
                       everything falls within these two points...

                       - Either the parent is critically unbalanced after
                         deletion or,
                       - It is completely in balance ...

                       Either case signals a change in tree height. For example:

                                        16
                                      ( +1 )
                                8               32
                              ( 0 )           ( -1 )
                                            24
                                           ( 0 )

                       If we remove node 24, 32's balance becomes 0. The tree
                       has changed height and as such, we need to trace to the
                       root, in order to bring it's balance down.

                       If we remove node 8, 16's balance becomes 2 - critically
                       right, which triggers a tree rebalance (rotation left).
                       This results in a change in tree height.

                       There is no need to trace to the root node unless there
                       has been a change of height elsewhere in the tree.

                       A change of height is only ever signalled by a parent
                       balance becoming BALANCED or LEFT/RIGHT_CRITICAL. So,
                       if the balance is neither balanced or critical, we can
                       escape from the trace :). */

                    if (trace_mode)
                        return;

                    break;
                }
            }

            /* Do we have anymore paths up or, is origin targetting the
               root node? */
            if (iterator)
            {
                /* Yep, We've got parents! Origin isn't targetted at root.
                   Adjust the parent's balance and continue tracing. */
                iterator->nBal +=
                    (multiplier * trace_constants[rel_cache]);

                origin = iterator;
                iterator = iterator->nLinks[PARENT];
            }
            else
                break;
        }
    }
}

/* -- init_node
 *
 * This call exists to initialize nodes, to make them
 * appropriate for tree insertion.
 *
 * Parameters:
 *   #1 - node, avl_hdr_t *
 *        Pointer to a valid avl_hdr_t structure.
 *        If this is null, the function will
 *        simply report failure.
 *
 *   #2 - key, unsigned long int.
 *        The identifier for the new node.
 *
 * Returns:
 *   On successful creation, this function will
 *   return a pointer to a new node.
 *
 *   On failure, it will return (avl_hdr_t *)0.
 *
 * --
*/
avl_hdr_t
*init_node (avl_hdr_t *node, unsigned long int key)
{
    if (node)
    {
        node->nLinks[LEFT] = (avl_hdr_t *)0;
        node->nLinks[RIGHT] = (avl_hdr_t *)0;
        node->nLinks[PARENT] = (avl_hdr_t *)0;

        node->nKey = key;
        node->nBal = BALANCED;
        node->nRel = INIT_CHILD;
    }
    return node;
}

/* -- insert_node
 *
 * This call will insert the node passed by the second
 * parameter, into the tree pointed to by the first.
 *
 * Please note, that the first paremeter is a POINTER
 * to a POINTER. This is so, if needed, the function can
 * update the 'tree' pointers target.
 *
 * Parameters:
 *  #1 - tree, avl_hdr_t **.
 *       A pointer to the root node of the tree.
 *       The target of the pointer may be NULL.
 *       (No existing root node in tree).
 *
 *  #2 - node, avl_hdr_t *
 *       Pointer to the node destined to inhabit the
 *       the tree. Nodes scheduled for insertion MUST
 *       be initialized. Inserting an uninitialized node
 *       will result in undefined behaviour.
 *       (Read: Initialize the nodes, or your program will
 *       crash faster than Windows 95...)
 *
 * Returns:
 *  On successful insertion, a pointer to the node inserted
 *  will be returned;
 *
 *  (avl_hdr_t *)0 will be returned in the case of
 *  key collision.
 *
 * Example of use:
 *
 *  avl_hdr_t *my_new_tree = (avl_hdr_t *)0;
 *  avl_hdr_t *a_node = create_node (0xCAFEBABE);
 *
 *  if (insert_node (&my_new_tree, a_node) == (avl_hdr_t *)0)
 *  {
 *    free (a_node);
 *    ... do error stuff ...
 *  }
 *
 * --
*/
avl_hdr_t
*insert_node (avl_hdr_t **tree, avl_hdr_t *node)
{
    /* Safety check, ensure the node being inserted
       is initialized BALANCED. Failure to start balanced
       will break tree integrity. */
    node->nBal = BALANCED;

    if (!*tree)
        *tree = node;
    else
    {
        avl_hdr_t *iterator = *tree;
        unsigned char tracker = 0;

        /* Iterate down the tree, searching for
           an appropriate insertion point. */
        while (1)
        {
            /* Which direction? Left or Right? */
            tracker = node->nKey > iterator->nKey;

            /* Gah! We've found a duplicate! */
            if (iterator->nKey == node->nKey)
                return (avl_hdr_t *)0;
            else
            if (iterator->nLinks[tracker])
                iterator = iterator->nLinks[tracker];
            else
            {
                /* Perfect, we've found the node a home! */
                iterator->nLinks[tracker] = node;
                break;
            }
        }

        /* Specify the node's relationship to it's parent. */
        node->nRel = tracker;
        node->nLinks[PARENT] = iterator;

        /* Update the parent's balance. */
        iterator->nBal += trace_constants[tracker];

        /* Trace up the tree, checking for imbalance. */
        trace (tree, iterator, INSERTION);
    }
    return node;
}

/* -- locate_node
 *
 * This call will attempt to locate the node identified
 * by the key provided.
 *
 * Parameters:
 *  #1 - tree, avl_hdr_t **
 *       A pointer to the root node of the tree.
 *       The target of the pointer may be NULL.
 *
 *  #2 - key, unsigned long int
 *       Idenfitier of that evasive, evil node...
 *       (Where in the world is Carmen Sandiego?)
 *
 * Returns:
 *  On success, a pointer to the now-discovered node is
 *  returned.
 *
 *  In all other cases, (avl_hdr_t *)0 is returned.
 *
 * --
*/
avl_hdr_t
*locate_node (avl_hdr_t **tree, unsigned long int key)
{
    if (*tree)
    {
        avl_hdr_t *iter = *tree;
        unsigned char tracker;

        while (1)
        {
            tracker = key > iter->nKey;

            if (iter->nKey == key)
                return iter;
            else if (iter->nLinks[tracker])
                iter = iter->nLinks[tracker];
            else
                return (avl_hdr_t *)0;
        }
    }
    return (avl_hdr_t *)0;
}

/* -- delete_node
 *
 * This call will remove the node identified from the
 * specified tree. Note, that this function may NOT release
 * the node you pass, due to the details of how this function
 * works.
 *
 * This call does NOT take responsibility for the deallocation
 * of the desired node. This makes it possible to use this call
 * to transfer nodes between trees.
 *
 * Parameters:
 *  #1 - tree, avl_hdr_t **
 *       A pointer to the root node of the tree.
 *       The target of the pointer may be NULL.
 *
 *  #2 - node, avl_hdr_t *
 *       Pointer to the node scheduled for deletion.
 *       This can only point to null in the case that the tree
 *       points to NULL. In any other case, this must NOT
 *       be zeroed.
 *
 * Returns:
 *  pointer to a node containing the data of the node removed.
 *
 *  For those anxious, no, this does not return a copy or have
 *  anything to do with wasting memory :). It is simply a side
 *  effect of implementation detail.
 *
 * --
*/
avl_hdr_t
*delete_node (avl_hdr_t **tree, avl_hdr_t *node)
{
    /* Does the tree consist of only a leaf? */
    if (*tree == node &&
        !node->nLinks[LEFT] && !node->nLinks[RIGHT])
    {
        /* Yes! It must be node. */
        /* Ensure the tree is DEAD! :) */
        *tree = (avl_hdr_t *)0;
    }
    else
    {
        /* In order to successfully remove the node from the tree,
           we need to understand precisely what type of node it is! */
        unsigned long int node_classification = 0;

        /* Does it have left/right children? */
        if (node->nLinks[LEFT])
            node_classification++;

        if (node->nLinks[RIGHT])
            node_classification++;

        switch (node_classification)
        {
            /* It has both, therefore it must be a Subtree. */
            case SUBTREE:
            {
                /* Find an appropriate successor for the subtree
                   node */
                avl_hdr_t *iterator = node->nLinks[LEFT];

                while (iterator->nLinks[RIGHT])
                    iterator = iterator->nLinks[RIGHT];

                /* Cache the subtree's key. */
                node_classification = node->nKey;

                /* Is the successor a parent? */
                if (iterator->nLinks[LEFT])
                {
                    /* Make the successor the new subtree root,
                       This is an illusion. All we are doing is
                       making the subtree node APPEAR as the
                       successor. */
                    node->nKey = iterator->nKey;

                    /* Make the successor a clone of it's child. */
                    iterator->nKey = iterator->nLinks[LEFT]->nKey;
                    /* Make the node lighter on the left, since we removed
                       the successor's child (In an illusionist way...) */
                    iterator->nBal++;

                    /* Prune the node once occupied by the successor's
                       left child from the tree. */
                    iterator = iterator->nLinks[LEFT];
                    iterator->nLinks[PARENT]->nLinks[LEFT] = (avl_hdr_t *)0;

                    /* And, make it appear to be the node we originally
                       wanted to remove. */
                    iterator->nKey = node_classification;
                }
                else
                {
                    /* Nope, the successor is a loner who spends all his
                       time in an apartment watching subpar sitcoms on
                       FOX... So, let's save him from his misery and
                       remove him from the tree :) */

                    /* Since we are removing this node, lighten it's
                       load on it's parent! */
                    iterator->nLinks[PARENT]->nBal +=
                        trace_constants[!iterator->nRel];

                    /* And, let it's parent know that we have subtly
                       killed once of it's children ... ;) */
                    iterator->nLinks[PARENT]->
                        nLinks[iterator->nRel] = (avl_hdr_t *)0;

                    /* BUT! In a crazy turn of events, The successor finds
                       his faith and, in the last moments of sweet cyber
                       life becomes King of the Subtree!. As such,
                       he know inhabits the node we intended to delete. */
                    node->nKey = iterator->nKey;

                    /* But... hey! We still have to return a node, that
                       appears as it was the node we ORIGINALLY wanted
                       to remove. So, let's move the former King of the
                       Subtree into the loner's apartment. And, return
                       him! */
                    iterator->nKey = node_classification;
                }

                /* Make iterator node's target, so we can uptrace properly. */
                node = iterator;
            }
            break;

            case BRANCH:
            {
                /* We need to discover which child of the branch
                   will succeed the branch itself. */
                unsigned char sibling = node->nBal > BALANCED;

                /* The successor takes the branch's place. As such,
                   it shares it's relationship. */
                node->nLinks[sibling]->nRel =
                    node->nRel;

                /* As well as parent... (OH, the scandel!) */
                node->nLinks[sibling]->nLinks[PARENT] =
                    (*node).nLinks[PARENT];

                /* However, it is possible that the branch was
                   in fact the root ndoe of the tree. */
                if (node->nLinks[PARENT])
                {
                    /* If not, we need to link the parent to it's
                       new branchy child. */
                    node->nLinks[PARENT]->
                        nLinks[node->nRel] =
                            node->nLinks[node->nBal > BALANCED];

                    /* And, alert it that we savagely murdered the
                       child THAT was there ... */
                    node->nLinks[PARENT]->nBal +=
                        trace_constants[!node->nRel];
                }
                else
                    /* If it is, we simply make the successor root. */
                    *tree = node->nLinks[node->nBal > BALANCED];
            }
            break;

            case LEAF:
            {
                /* It's a leaf node! The simplest, friendliest node of all!*/
                if (node->nLinks[PARENT])
                {
                    /* If the leaf node isn't root, we simply make it's
                       parent aware of the fact that we killed it's only
                       child ... */
                    node->nLinks[PARENT]->
                        nLinks[node->nRel] = (avl_hdr_t *)0;

                    /* And, let it know that their isn't so much weight on
                       the parent's bank account anymore ... :P */
                    node->nLinks[PARENT]->nBal +=
                        trace_constants[!node->nRel];
                }
                else
                    /* If the leaf WAS root, then we have officially killed
                       the tree. */
                    *tree = (avl_hdr_t *)0;
            }
            break;
        }
        /* Finally, we need to traverse up from the point of true deletion,
           bringing the tree back into balance if necessary. */
        trace (tree, node->nLinks[PARENT], DELETION);
    }

    /* And, return the node (which will always appear as the node we wanted to
       delete in the first place ...) for destruction. */
    return node;
}

#ifdef _test_avl /* _test_avl */
#include <stdio.h>
#include <stdlib.h>

    int
    main (void)
    {
        avl_hdr_t *avl_tree = (avl_hdr_t *)0;
        avl_hdr_t *avl_nodes[_test_avl];
        unsigned long int avl_inserts, avl_node_c = 0, avl_fail_count = 0;

        printf(" !! Testing init_node (...) / insert_node (...)\r\n");
        for (avl_inserts = 0; avl_inserts < _test_avl; avl_inserts++)
        {
            printf ("\tCreating node(0x%lX) ... ", avl_inserts);

            avl_nodes[avl_inserts] =
                init_node ((avl_hdr_t *)malloc (sizeof (avl_hdr_t)),
                avl_inserts);

            if (avl_nodes[avl_inserts])
            {
                printf ("OKAY!\r\n");
                printf ("\t + Inserting node(0x%lX) ... ", avl_inserts);

                if (insert_node (&avl_tree, avl_nodes[avl_inserts]))
                {
                    avl_node_c++;
                    printf ("OKAY!\r\n");
                }
                else
                {
                    printf ("FAIL!\r\n");
                    avl_fail_count++;
                }
            }
            else
            {
                printf ("FAIL!\r\n");
                avl_fail_count++;
            }
        }

        printf("\r\n !! Testing locate_node (...)\r\n");
        for (avl_inserts = _test_avl; avl_inserts > 0; avl_inserts--)
        {
            printf ("\tResolving node(0x%lX) ... ", avl_inserts - 1);
            avl_nodes[_test_avl - avl_inserts] =
                locate_node (&avl_tree, avl_inserts - 1);

            if (avl_nodes[_test_avl - avl_inserts])
            {
                printf ("OKAY!\r\n");
                printf ("\t + Verifying node(0x%lX) ... ", avl_inserts - 1);
                if (avl_nodes[_test_avl - avl_inserts]->nKey == avl_inserts - 1)
                    printf ("OKAY!\r\n");
                else
                {
                    printf ("FAIL!\n");
                    avl_fail_count++;
                }
            }
            else
            {
                printf ("FAIL!\r\n");
                avl_fail_count++;
            }
        }


        #ifdef _remove_all_max /* _remove_all_max */
        printf("\r\n !! Testing delete_node (...)\r\n");

        for (avl_inserts = _test_avl; avl_inserts > 0; avl_inserts--)
        {
            printf ("\tPopping node(0x%lX) ... ", avl_inserts - 1);
            avl_nodes[_test_avl - avl_inserts] =
                delete_node (&avl_tree, avl_nodes[_test_avl - avl_inserts]);

            if (avl_nodes[_test_avl - avl_inserts])
            {
                printf ("OKAY!\r\n");
                printf ("\t + Verifying node(0x%lX) ... ", avl_inserts - 1);
                if (avl_nodes[_test_avl - avl_inserts]->nKey != avl_inserts - 1)
                {
                    printf ("FAIL!\r\n");
                    avl_fail_count++;
                }
                else
                {
                    printf ("OKAY!\r\n");
                    printf ("\t + Releasing node(0x%lX) ... ", avl_inserts - 1);
                    free (avl_nodes[_test_avl - avl_inserts]);
                    avl_nodes[_test_avl - avl_inserts] = (avl_hdr_t *)0;
                    printf("DONE!\r\n");
                }
            }
            else
            {
                printf ("FAIL!\r\n");
                avl_fail_count++;
            }
        }
        #endif /* _remove_all_max */

        printf ("\r\n !! Ensuring all assets have been released ... ");

        while (avl_tree)
            free (delete_node (&avl_tree, avl_tree));

        printf ("CLEAN!\r\n");
        printf (" ?? Statistics : [ %lu insetions : %lu failure(s) ].\r\n",
                avl_node_c, avl_fail_count);

        return EXIT_SUCCESS;
    }

#endif /* _test_avl */

